#if os(iOS) || os(watchOS) || os(tvOS)
  import UIKit
#else
  import Foundation
#endif

import Tailor

/**
 A value type struct, it conforms to the Mappable protocol so that it can be instantiated with JSON
 */
public struct Item: Mappable {

  /**
   An enum with all the string keys used in the view model
   */
  public enum Key: String {
    case Index
    case Identifier
    case Title
    case Subtitle
    case Image
    case Type
    case Kind
    case Action
    case Meta
    case Children
    case Relations
    case Size
    case Width
    case Height

    var string: String {
      return rawValue.lowercaseString
    }
  }

  /// The index of the Item when appearing in a list, should be computed and continuously updated by the data source
  public var index = 0
  /// An optional identifier for your data
  public var identifier: Int?
  /// The main representation of the Item
  public var title = ""
  /// Supplementary information to the Item
  public var subtitle = ""
  /// A visual representation of the Item, usually a string URL or image name
  public var image = ""
  /// Determines what kind of UI should be used to represent the Item
  public var kind: String = ""
  /// A string representation of what should happens when a Item is tapped, usually a URN or URL
  public var action: String?
  /// The width and height of the view model, usually calculated and updated by the UI component
  public var size = CGSize(width: 0, height: 0)
  /// A collection of items used for composition
  public var children: [[String : AnyObject]] = []
  /// A key-value dictionary for any additional information
  public var meta = [String : AnyObject]()
  /// A key-value dictionary for related view models
  public var relations = [String : [Item]]()

  /// A dictionary representation of the view model
  public var dictionary: [String : AnyObject] {
    var dictionary: [String: AnyObject] = [
      Key.Index.string : index,
      Key.Kind.string : kind,
      Key.Size.string : [
        Key.Width.string : size.width,
        Key.Height.string : size.height
      ]
    ]

    if !title.isEmpty { dictionary[Key.Title.string] = title }
    if !subtitle.isEmpty { dictionary[Key.Subtitle.string] = subtitle }
    if !image.isEmpty { dictionary[Key.Image.string] = image }
    if !meta.isEmpty { dictionary[Key.Meta.string] = meta }

    if let identifier = identifier {
      dictionary[Key.Identifier.string] = identifier
    }

    if let action = action {
      dictionary[Key.Action.string] = action
    }

    dictionary[Key.Children.string] = children

    var relationItems = [String : [[String : AnyObject]]]()

    relations.forEach { key, array in
      if relationItems[key] == nil { relationItems[key] = [[String : AnyObject]]() }
      array.forEach { relationItems[key]?.append($0.dictionary) }
    }

    if !relationItems.isEmpty {
      dictionary[Key.Relations.string] = relationItems
    }

    return dictionary
  }

  // MARK: - Initialization

  /**
   Initialization a new instance of a Item and map it to a JSON dictionary

   - parameter map: A JSON dictionary
   */
  public init(_ map: [String : AnyObject]) {
    index    <- map.property(.Index)
    identifier = map.property(.Identifier)
    title    <- map.property(.Title)
    subtitle <- map.property(.Subtitle)
    image    <- map.property(.Image)
    kind     <- map.property(.Type) ?? map.property(.Kind)
    action   <- map.property(.Action) ?? nil
    meta     <- map.property(.Meta)
    children = map[.Children] as? [[String : AnyObject]] ?? []

    if let relation = map[.Relations] as? [String : [Item]] {
      relations = relation
    }

    if let relations = map[.Relations] as? [String : [[String : AnyObject]]] {
      var newRelations = [String : [Item]]()
      relations.forEach { key, array in
        if newRelations[key] == nil { newRelations[key] = [Item]() }
        array.forEach { newRelations[key]?.append(Item($0)) }

        self.relations = newRelations
      }
    }

    size = CGSize(
      width:  ((map[.Size] as? [String : AnyObject])?[.Width] as? Int) ?? 0,
      height: ((map[.Size] as? [String : AnyObject])?[.Height] as? Int) ?? 0
    )
  }

  /**
   Initialization a new instance of a Item and map it to a JSON dictionary

   - parameter title: The title string for the view model, defaults to empty string
   - parameter subtitle: The subtitle string for the view model, default to empty string
   - parameter image: Image name or URL as a string, default to empty string
   */
  public init(identifier: Int? = nil, title: String = "", subtitle: String = "", image: String = "", kind: StringConvertible = "", action: String? = nil, size: CGSize = CGSize(width: 0, height: 0), meta: [String : AnyObject] = [:], relations: [String : [Item]] = [:]) {
    self.identifier = identifier
    self.title = title
    self.subtitle = subtitle
    self.image = image
    self.kind = kind.string
    self.action = action
    self.size = size
    self.meta = meta
    self.relations = relations
  }

  /**
   Initialization a new instance of a Item and map it to a JSON dictionary

   - parameter title: The title string for the view model, defaults to empty string
   - parameter subtitle: The subtitle string for the view model, default to empty string
   - parameter image: Image name or URL as a string, default to empty string
   */
  public init(identifier: Int? = nil, title: String = "", subtitle: String = "", image: String = "", kind: StringConvertible = "", action: String? = nil, size: CGSize = CGSize(width: 0, height: 0), meta: Mappable, relations: [String : [Item]] = [:]) {
    self.init(identifier: identifier, title: title, subtitle: subtitle, image: image, kind: kind, action: action,
              size: size, meta: meta.metaProperties, relations: relations)
  }

  // MARK: - Helpers

  /**
   A generic convenience method for resolving meta attributes

   - parameter key: String
   - parameter defaultValue: A generic value that works as a fallback if the key value object cannot be cast into the generic type

   - returns: A generic value based on `defaultValue`, it falls back to `defaultValue` if type casting fails
   */
  public func meta<T>(key: String, _ defaultValue: T) -> T {
    return meta[key] as? T ?? defaultValue
  }

  /**
   A generic convenience method for resolving meta attributes

   - parameter key: String
   - parameter type: A generic type used for casting the meta property to a specific value or reference type

   - returns: An optional generic value based on `type`
   */
  public func meta<T>(key: String, type: T.Type) -> T? {
    return meta[key] as? T
  }

  /**
   A generic convenience method for resolving meta instance

   - returns: A generic meta instance based on `type`
   */
  public func metaInstance<T: Mappable>() -> T {
    return T(meta)
  }

  /**
   A convenience lookup method for resolving view model relations

   - parameter key: String
   - parameter index: The index of the object inside of `self.relations`
   */
  public func relation(key: String, _ index: Int) -> Item? {
    if let items = relations[key] where index < items.count {
      return items[index]
    } else {
      return nil
    }
  }

  /**
   A method for mutating the kind of a view model

   - parameter kind: A StringConvertible object
   */
  public mutating func update(kind kind: StringConvertible) {
    self.kind = kind.string
  }
}

/**
 A collection of Item Equatable implementation
 - parameter lhs: Left hand collection of Items
 - parameter rhs: Right hand collection of Items
 - returns: A boolean value, true if both Item are equal
 */
public func ==(lhs: [Item], rhs: [Item]) -> Bool {
  var equal = lhs.count == rhs.count

  if !equal { return false }

  for (index, item) in lhs.enumerate() {
    if item != rhs[index] { equal = false; break }
  }

  return equal
}

/**
 A collection of Item Equatable implementation to see if they are truly equal
 - parameter lhs: Left hand collection of Items
 - parameter rhs: Right hand collection of Items
 - returns: A boolean value, true if both Item are equal
 */
public func ===(lhs: [Item], rhs: [Item]) -> Bool {
  var equal = lhs.count == rhs.count

  if !equal { return false }

  for (index, item) in lhs.enumerate() {
    if !(item === rhs[index]) {
      equal = false
      break
    }
  }

  return equal
}

/**
 Item Equatable implementation
 - parameter lhs: Left hand Item
 - parameter rhs: Right hand Item

 - returns: A boolean value, true if both Item are equal
 */
public func ==(lhs: Item, rhs: Item) -> Bool {
  return lhs.identifier == rhs.identifier &&
    lhs.title == rhs.title &&
    lhs.subtitle == rhs.subtitle &&
    lhs.image == rhs.image &&
    lhs.kind == rhs.kind &&
    lhs.action == rhs.action &&
    (lhs.meta as NSDictionary).isEqualToDictionary(rhs.meta) &&
    compareRelations(lhs, rhs)
}

/**
 Check if Item's are truly equal by including size in comparison

 - parameter lhs: Left hand Item
 - parameter rhs: Right hand Item

 - returns: A boolean value, true if both Item are equal
 */
public func ===(lhs: Item, rhs: Item) -> Bool {
  let equal = lhs.identifier == rhs.identifier &&
    lhs.title == rhs.title &&
    lhs.subtitle == rhs.subtitle &&
    lhs.image == rhs.image &&
    lhs.kind == rhs.kind &&
    lhs.action == rhs.action &&
    lhs.size == rhs.size &&
    (lhs.meta as NSDictionary).isEqualToDictionary(rhs.meta) &&
    compareRelations(lhs, rhs)

  return equal
}

/**
 A reverse Equatable implementation for comparing Item's
 - parameter lhs: Left hand Item
 - parameter rhs: Right hand Item

 - returns: A boolean value, false if both Item are equal
 */
public func !=(lhs: Item, rhs: Item) -> Bool {
  return !(lhs == rhs)
}

func compareRelations(lhs: Item, _ rhs: Item) -> Bool {
  guard lhs.relations.count == rhs.relations.count else {
    return false
  }

  var equal = true

  for (key, value) in lhs.relations {
    guard let rightValue = rhs.relations[key] where value == rightValue
      else { equal = false; break }
  }

  return equal
}
